# This file is part of DesktopDig
# Copyright (C) 2002 Daniel Naber <daniel.naber@t-online.de>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import codecs
import commands
import os
import re
import stat
import string
import zipfile

class Converter:
	"""Convert many different files types to text and extract meta information
	if possible."""
	
	def __init__(self, filename, desc_words, default_encoding):
		self.filename = filename
		suffix_match = re.compile(".*\.(.*)").match(self.filename)
		if suffix_match:
			self.suffix = suffix_match.group(1).lower()
		else:
			self.suffix = None
		self.desc_words = desc_words
		self.default_encoding = default_encoding
		return
		
	def getDocumentSize(self):
		size = os.stat(self.filename)[stat.ST_SIZE]
		return size

	def getDocumentInfo(self):
		filesize = self.getDocumentSize()
		(err, title, author, desc, content) = self.convert()
		date = os.stat(self.filename).st_mtime
		return (err, filesize, date, title, author, desc, content)

	def removeXMLComments(self, s):
		start_pos = s.find("<!--")
		while start_pos != -1:
			end_pos = s.find("-->", start_pos)
			if end_pos != -1:
				s = "%s%s" % (s[0:start_pos], s[end_pos+3:])
			start_pos = s.find("<!--")
		return s

	def removeXMLMarkup(self, s, replace_with_space):
		s = re.compile("<!--.*?-->", re.DOTALL).sub('', s)
		repl = ''
		if replace_with_space:
			repl = ' '
		s = re.compile("<[^>]*>", re.DOTALL).sub(repl, s)
		return s

	def getEncodedFile(self, content):
		"""Encode the file according to its <meta> tag. Otherwise use
		the default encoding."""
		#print "##%s (%s)" % (self.filename, type(content))
		if not content:
			return
		content = self.removeXMLComments(content)
		# TODO: this should not only work for text/html, but also
		# for e.g. text/xml, so this belongs somewhere else.
		meta_regex = re.compile("<meta([^>]+)>", re.DOTALL|re.IGNORECASE)
		http_regex = re.compile("http-equiv\s*=\s*[\"']Content-Type[\"']", re.DOTALL|re.IGNORECASE)
		value_regex = re.compile("content\s*=\s*[\"']text/html;\s*charset\s*=\s*(.*?)[\"']", re.DOTALL|re.IGNORECASE)
		meta_matches = meta_regex.findall(content)
		enc = None
		for meta_match in meta_matches:
			if http_regex.search(meta_match):
				charset_match = value_regex.search(meta_match)
				if charset_match:
					enc = charset_match.group(1)
					print "Loading '%s' with encoding '%s'" % (self.filename, enc)
					# errors are catched in Indexer.indexDirectories():
					content = content.decode(enc)
					break
		if not enc:
			# special case is necessary for utf-16 since this isn't
			# found by the regexp above if the file is not loaded as utf-16:
			try:
				f = codecs.open(self.filename, 'r', 'utf-16')
				content = f.read()
				f.close()
				print "Loading '%s' with encoding 'utf-16'" % self.filename
			except UnicodeError:
				print "Loading '%s' with default encoding '%s'" % (self.filename, self.default_encoding)
				f = codecs.open(self.filename, 'r', self.default_encoding)
				content = f.read()
				f.close()
		return content
		
	def convert(self):
		err = None
		title = None
		author = None
		desc = None
		content = None

		##### Plain text:
		if self.suffix in ('txt', 'ascii', 'asc', 'py', 'pl', 'cc', 'cpp', 'c', 'h', 'java'):
			content = self.loadDefaultEncoding(self.filename)

		##### HTML, ...:
		elif self.suffix in ('html', 'htm', 'php', 'php3', 'php4', 'xml'):
			content = self.load(self.filename)
			content = self.getEncodedFile(content)
			if not content:
				return ("Empty file", None, None, None, None)
			content = self.removeXMLComments(content)
			title_match = re.compile("<title>(.*?)</title>", re.DOTALL|re.IGNORECASE).search(content)
			if title_match:
				title = title_match.group(1)
			# fixme: allow ["'] as delimiter:
			# fixme: don't care about order of attributes
			author_match = re.compile('<meta\s+name\s*=\s*"author" content\s*=\s*"([^"]*)"', re.DOTALL|re.IGNORECASE).search(content)
			if author_match:
				author = author_match.group(1)
			desc_match = re.compile('<meta\s+name\s*=\s*["\']description["\'] content\s*=\s*"([^"]*)"', re.DOTALL|re.IGNORECASE).search(content)
			if desc_match:
				desc = desc_match.group(1)
			else:
				# TODO: not efficient for very large HTML files...
				body_match = re.compile("<body[^>]*>(.*)</body>", re.DOTALL|re.IGNORECASE).search(content)
				desc_tmp =  ''
				if body_match:
					desc_tmp = body_match.group(1)
				else:
					# a HTML file without <body>..</body>
					desc_tmp = content
				desc_tmp = re.compile("<.*?>", re.DOTALL).sub('', desc_tmp)
				desc_terms = re.split("\s+", desc_tmp)
				desc = string.join(desc_terms[0:self.desc_words], ' ')
				# fixme: this 'desc' needs to be marked as "generated", otherwise
				# it's indexed and the terms are counted twice, which
				# increases the file's ranking:
				if len(desc_terms) > self.desc_words:
					desc = "%s..." % desc

		##### KWord:
		elif self.suffix in ('kwd',):
			try:
				zip = zipfile.ZipFile(self.filename, "r")
				content = zip.read("maindoc.xml").decode('utf-8')
				content = self.removeXMLMarkup(content, 0)
				docinfo = zip.read("documentinfo.xml")
				title_match = re.compile("<about.*?<title>(.*?)</title>", re.DOTALL|re.IGNORECASE).search(docinfo)
				if title_match:
					title = title_match.group(1)
				# TODO: extract more meta information
			except zipfile.BadZipfile:
				content = None
				err = "%s: Old KOffice (< 1.2) file format not supported." % self.filename
				# TODO: support them (it's just a tar.gz)

		##### OpenOffice:
		elif self.suffix in ('sxw',):
			try:
				zip = zipfile.ZipFile(self.filename, "r")
				content = zip.read("content.xml").decode('utf-8')
				content = self.removeXMLMarkup(content, 1)
				docinfo = zip.read("meta.xml").decode('utf-8')
				#print docinfo.encode('latin1')
				title_match = re.compile("<dc:title>(.*?)</dc:title>", re.DOTALL|re.IGNORECASE).search(docinfo)
				if title_match:
					title = title_match.group(1)
				desc_match = re.compile("<dc:description>(.*?)</dc:description>", re.DOTALL|re.IGNORECASE).search(docinfo)
				if desc_match:
					desc = desc_match.group(1)
				author_match = re.compile("<dc:creator>(.*?)</dc:creator>", re.DOTALL|re.IGNORECASE).search(docinfo)
				if author_match:
					author = author_match.group(1)
			except zipfile.BadZipfile:
				content = None
				err = "%s: Supposed ZIP file could not be opened." % self.filename

		##### PDF:
		elif self.suffix in ('pdf',):
			cmd = "pdftotext -enc UTF-8 %s -" % commands.mkarg(self.filename)
			(status, content) = commands.getstatusoutput(cmd)
			content = content.decode('utf-8')
			if not status == 0:
				err = "Could not successfully execute '%s'" % cmd
				content = None
		# TODO: lyx
		# TODO: ps (with ps2ascii? fails on my language paper!), zip, tgz, doc, ...
		# TODO: docbook
		# TODO: use external scripts if available (e.g. "application/word" from "converter" sub directory??)
		#  TODO: make these external (as an example?)
		# TODO: maildir mails!

		##### Everything else:
		else:
			# TODO: make this a non-error and take the file as if it was 
			# plain text?
			#content = None
			#err = "No converter found for suffix '%s' (%s)" % (self.suffix, self.filename)
			content = self.loadDefaultEncoding(self.filename)
		return (err, title, author, desc, content)

	def load(self, filename):
		f = open(filename)
		contents = f.read()
		f.close()
		return contents

	def loadDefaultEncoding(self, filename):
		f = codecs.open(self.filename, 'r', self.default_encoding)
		contents = f.read()
		f.close()
		return contents
