# This file is part of DesktopDig
# Copyright (C) 2002,2003 Daniel Naber <daniel.naber@t-online.de>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

# TODO:
# -support "bla*" -> terms = getTruncatedTerms(terms), then search like always, using terms
# -beedict 2.1 is going to have an configurable cache size

# self.table: This is the inverted index data structure (example with phrase_search = 1):
# 
# {
#	('term1', {
#				(doc_id1: (weight1, [pos_in_doc1])),
#				(doc_id2: (weight2, [pos_in_doc1, pos_in_doc2, pos_in_doc3]))
#			  }),
#	('term2', {
#				(doc_id1: (weight3, [pos_in_doc3]))
#			}),
# }
#
# If phrase_search = 0, there won't be a list of positions, instead there'll
# be an integer that counts the number of this terms' occurences in this doc.

import copy 
import os
import math
import re
import string
import sys
import time
import types

import Documents
import IgnoreTerms
import Tokenizer

from mx.BeeBase.BeeDict import BeeDict

import Set

true = 1
false = 0

class FullTextIndex:
	"""An experimental inverted index that doesn't use lists to save
	the document ids, but hash tables. This seems to speed up 
	removeDocument() greatly (0.6 sec instead of 1.3 seconds, and
	1.2sec instead of 2.9sec - i.e. it seems to scale much better!)"""

	def __init__(self, basedir, db_name_prefix, searchonly, min_length=2, 
			max_length=15, phrase_search=1, files_world_readable=0):
		self.basedir = basedir
		self.searchonly = searchonly
		self.min_length = min_length
		self.max_length = max_length
		self.phrase_search = phrase_search
		self.files_world_readable = files_world_readable
		self.db_name = os.path.join(self.basedir, db_name_prefix+"_terms_bee")
		self.db_words_name = os.path.join(self.basedir, db_name_prefix+"_words_bee")
		self.db_docs_name = os.path.join(self.basedir, db_name_prefix+"_docs_bee")
		self.table = {}		# the inverted index - bound to deedict
		self.table_cache = {}		# the inverted index - cache, i.e. in-memory
		self.docs_table = {}		# self.docs[id] = doc_object
		self.words_table = {}		# words_table[doc_id] = [word1, word2, ...]	(for update speedup)
		self.words_occ_table = {}	# words_occ_table[term] = #occurences_all_docs (for relevance ranking)
		self.count = 0
		self.stopwords = IgnoreTerms.IgnoreTerms(self.basedir, self.min_length, self.max_length)
		self.flush_count = 0
		self.docs_indexed = 0
		# for testing:
		self.timect = 0
		return

	def getIndexDocumentNames(self):
		l = []
		for doc in self.docs_table.values():
			l.append(doc.url)
		return l
	
	def countDocuments(self):
		return len(self.docs_table.keys())
		
	def cleanupIndex(self):
		"""Remove all documents from the index whose files don't exist 
		anymore."""
		for doc_id in self.docs_table.keys():
			filename = self.docs_table[doc_id].url
			if not os.path.exists(filename):
				print "File '%s' was removed" % self.docs_table[doc_id].url
				self.removeDocument(doc_id)
			#else:
			#	print "File '%s' is still there" % self.docs_table[doc_id].url
		return

	def bindData(self):
		if self.files_world_readable:
			os.umask(022)
		else:
			# for security reasons (we're indexing the user's
			# personal files!), set file's permissions to -rw-------:
			os.umask(077)
		self.table = BeeDict(self.db_name, min_recordsize=0, readonly=self.searchonly, \
			recover=0, autocommit=1, validate=0)
		self.docs_table = BeeDict(self.db_docs_name, min_recordsize=0, readonly=self.searchonly, \
			recover=0, autocommit=1, validate=0)
		if not self.searchonly:
			self.words_table = BeeDict(self.db_words_name, min_recordsize=0, readonly=self.searchonly, \
				recover=0, autocommit=1, validate=0)
		return

	def commitData(self):
		#print >> sys.stderr, "Known words = %d" % len(self.table.keys())
		#print >> sys.stderr, "Commiting results..."
		self.table.commit()
		self.words_table.commit()
		self.docs_table.commit()
		self.table.close()
		self.words_table.close()
		self.docs_table.close()
		return

	def deleteData(self):
		try:
			#print >> sys.stderr, "Deleting old data files..."
			os.remove(self.db_name + ".dat")
			os.remove(self.db_name + ".idx")
			os.remove(self.db_words_name + ".dat")
			os.remove(self.db_words_name + ".idx")
			os.remove(self.db_docs_name + ".dat")
			os.remove(self.db_docs_name + ".idx")
		except OSError:
			print >> sys.stderr, "No old data files found."
		return

	def __str__(self):
		"Print self for debugging."
		l = []
		for term in self.table.keys():
			subtable = self.table[term]
			lsub = []
			l.append("%s: %s" % (string.rjust(term, 15), subtable))
		return string.join(l, "\n")

	def toList(self):
		"Print self for test cases."
		l = []
		for term in self.table.keys():
			subtable = self.table[term]
			lsub = []
			for key in subtable.keys():
				# ignore the value, it's just always "1"
				positions = subtable[key]
				#print "#POS = %s" % positions
				lsub.append((key, positions))
			lsub.sort()
			l.append((term,lsub))
		l.sort()
		#print "L"+str(l)
		return l

	def addDocument(self, doc_id, text, title="__testcase__", 
			author="__testcase__", description="__testcase__",
			filename="__testcase__", size=0, date=0):
		#t1 = time.time()
		#todo: raise error if id exists already?
		self.count = self.count + 1
		tk = Tokenizer.Tokenizer()		# todo: how to use this as static??
		terms = []
		if title != "__testcase__":
			terms = tk.tokenize('%s %s %s %s' % \
				(title, author, description, text), ignore_quotes=1)
			#print terms
		else:
			terms = tk.tokenize(text, ignore_quotes=1)
		# A doc_id is automatically generated, unless the document is
		# updated, i.e. removed and added back. In the update case,
		# we re-use the old document id.
		if not doc_id:
			doc_id = self.addDocumentToCache(None, terms)
		else:
			self.addDocumentToCache(doc_id, terms)
		# modulo 100 speeds up indexing greatly, and uses only slightly more RAM:
		if self.count % 100 == 0:
			self.addDocumentToIndex(docs_so_far=self.count)
		doc = Documents.IndexedDocument(doc_id, filename, title=title, \
			author=author, description=description, size=size, date=date, \
			term_count=len(terms)	# TODO?: not 100% exact b/c of descr. etc.
			)
		if self.docs_table.has_key(doc_id):
			n = self.docs_table[doc_id].url
			# TODO: make this as error?
			print "*** Warning: Doc with ID %d already exists (%s) new=%s." % (doc_id, n, filename)
		self.docs_table[doc_id] = doc
		### commiting to disk, in order to save RAM:
		steps = 50
		if (self.count % steps == 0) and doc_id != 0:
			self.flush_count += 1
			#print "Flush %d..." % self.flush_count
			self.table.commit()
			self.words_table.commit()
		return doc_id

	def finishIndex(self):
		"""This *must* be called after changes to the index are finished. Otherwise
		a part of the information will not be saved."""
		self.addDocumentToIndex(docs_so_far=self.count, finish=1)
		return

	# TODO: move to some tools module:
	def getMax(self, l):
		maxi = None
		for elem in l:
			if maxi == None or elem > maxi:
				maxi = elem
		return maxi

	def addDocumentToCache(self, doc_id, terms):
		"""Save the inverted index in a normal Python hashtable which is in RAM.
		Accessing this (read and write) is very fast."""
		term_pos = 0
		if not doc_id:
			if self.docs_indexed == 0:
				# first run, find max id:
				doc_id = self.getMax(self.docs_table.keys())
				if not doc_id:
					doc_id = 0
				doc_id = doc_id + 1
				self.docs_indexed = doc_id
				#print "self.docs_indexed first = %d" % self.docs_indexed
			else:
				doc_id = self.docs_indexed
		self.docs_indexed = self.docs_indexed + 1
		### Save terms' positions:
		n_terms = []	# not ignored terms
		for term in terms:
			term = self.stopwords.acceptTerm(term)
			if not term:
				continue
			else:
				n_terms.append(term)
			if self.table_cache.has_key(term):
				subtable = self.table_cache[term]
				if subtable.has_key(doc_id):
					if self.phrase_search:
						subtable[doc_id].append(term_pos)
					else:
						subtable[doc_id] = subtable[doc_id]+1	# number of occurences
				else:
					if self.phrase_search:
						subtable[doc_id] = [term_pos]
					else:
						subtable[doc_id] = 1
			else:
				if self.phrase_search:
					self.table_cache[term] = {doc_id : [term_pos]}
				else:
					self.table_cache[term] = {doc_id : 1}
			term_pos = term_pos + 1
			### Save overall term occurences (to see what's a frequent term and what's not):
			if self.words_occ_table.has_key(term):
				self.words_occ_table[term] = self.words_occ_table[term] + 1
			else:
				self.words_occ_table[term] = 1
		### Save words per document, only once per document -> not cached:
		s = Set.Set()
		s.addList(n_terms)
		self.words_table[doc_id] = s.asList()
		return doc_id
		
	def addDocumentToIndex(self, docs_so_far, finish=0):
		"""Save the data from the cache to BeeDict, which looks like just another
		hashtable but it is stored on disk when commit() is called. This way we have
		a compromise between fast access (cache) and limited memory usage."""
		t1 = time.clock()
		table_cache = copy.deepcopy(self.table_cache)
		for term in table_cache:
			frequency = 1
			if docs_so_far > 0:
				frequency = self.words_occ_table[term] / float(docs_so_far)
			# higher values seem to lead to faster indexing?!:
			#limit = 3
			limit = 30
			#limit = 300
			if frequency > limit and not finish:
				# common term, ignore for now to speed things up,
				# unless this is the last run -- problem: this seems to slow down!!!
				#print "*%s -> %f" % (term, frequency)
				continue
			else:
				# add cached data to disk-based index:
				if self.table.has_key(term):
					subtable = self.table[term]
					subtable_cached = table_cache[term]
					for doc_id in subtable_cached.keys():
						#print "#%s" % doc_id
						if subtable.has_key(doc_id):
							l = subtable[doc_id]
							### fixme: is this 'if' needed?
							if type(l) is types.TupleType:
								#print "L=%s" % str(l)
								l[1].extend(subtable_cached[doc_id])
							else:
								l.extend(subtable_cached[doc_id])
							#del subtable_cached[doc_id]
							subtable[doc_id] = l	# required by beedict with commit()
						else:
							subtable[doc_id] = copy.deepcopy(subtable_cached[doc_id])
							#del subtable_cached[doc_id]
					self.table[term] = subtable   # required by beedict with commit()
				else:
					#print "*"+str(table_cache[term])
					self.table[term] = copy.deepcopy(table_cache[term])
				del self.table_cache[term]
		#print "addDocumentToIndex end %.2f (finish=%d)" % (time.time()-t1, finish)
		self.timect = self.timect + (time.clock()-t1)
		return

	def removeDocument(self, doc_id):
		if not self.words_table.has_key(doc_id):
			print "*** Error in removeDocument(): ID %d not found" % doc_id
			return
		if not self.docs_table.has_key(doc_id):
			print "*** Error (2) in removeDocument(): ID %d not found" % doc_id
			return
		del self.docs_table[doc_id]
		# Optimization: don't iterate over all terms, but only over those in
		#  the doc that should be deleted. I.e. this sues a list (set) of words
		#  that are in each doc (this is done in addDocument()).
		for term in self.words_table[doc_id]:
			if self.table.has_key(term):
				subtable = self.table[term]
				if subtable.has_key(doc_id):
					del subtable[doc_id]
					self.table[term] = subtable		# required by BeeBase
			# fixme?: this confuses BeeBase?:
			#if len(subtable.keys()) == 0:
			#	del self.table[term]
			#else:
			#	print "*** key '%s' not found" % term
		return
	
	def updateDocument(self, doc_id, new_text, 
			title='__testcase__', author='__testcase__', desc='__testcase__',
			filename='__testcase__', size=0, date=0):
		"""Update the document by removing and adding it again."""
		self.removeDocument(doc_id)
		self.addDocument(doc_id, new_text, title, author, desc, filename, 
			size, date)
		return

	def searchTerm(self, term):
		"""Returns a list of tuples (doc_id, relevancy) for all documents that contain the term.
		The term is not normalized, this is up to the caller."""
		if not self.table.has_key(term):
			return []
		else:
			l = []
			for doc_id in self.table[term].keys():
				if self.phrase_search:
					weight = self.table[term][doc_id][0]
				else:
					weight = self.table[term][doc_id]
				l.append((doc_id, weight))
			return l

	def searchPhrase(self, terms):
		"""Returns list of tuples (doc_id, relevancy) for documents that contain the phrase.
		The terms are not normalized (e.g. stopwords are not removed), this is up to the caller.
		fixme: throw error if not phrase_search """
		i = 0
		results = []
		positions = {}
		for term in terms:
			i = i + 1
			if not self.table.has_key(term):
				return []
			else:
				if i == 1:
					# first term in phrase, fill the dict with start information (first match):
					positions = copy.deepcopy(self.table[term])
				else:
					for doc_id in self.table[term].keys():
						new_positions = copy.deepcopy(self.table[term])
						for tmp_doc_id in positions.keys():
							if new_positions.has_key(tmp_doc_id):
								# okay, the current term is also found in this doc...
								# so far this is just like an AND search, but now we also
								# check the positions:
								new_pos = new_positions[tmp_doc_id][1]		# at [0] is the weight
								# positions of the relevant term in the previously found doc:
								prev_pos = positions[tmp_doc_id]
								found = false
								if prev_pos:
									prev_pos = prev_pos[1]
									for tmp_prev_pos in prev_pos:
										try:
											# TODO: add proximity search here
											x = new_pos.index(tmp_prev_pos + (i-1))
											found = true	# only if index() had thrown no exception
										except ValueError:
											pass
								if not found:
									positions[tmp_doc_id] = None
							else:
								# current term not found
								positions[tmp_doc_id] = None
		for tmp_doc_id in positions.keys():
			tmp_weight = 0
			if positions[tmp_doc_id]:		# because None is set above
				for term in terms:
					tmp_weight += self.table[term][tmp_doc_id][0]
				results.append((tmp_doc_id, tmp_weight))
		return results

	def calculateRanking(self):
		"""Assign each (term, doc_id) tuple a ranking, depending on the number
		of occurences of the term in the document and in all documents."""
		docs_count = len(self.words_table.keys())
		i = 0
		for term in self.table.keys():
			if i % 100 == 0:
				# TODO: normalize this, i.e. make it work towards a 100% limit
				sys.stdout.flush()
				print ".",
			i = i + 1
			subtable = self.table[term]
			term_count = len(subtable.keys())	# in how many of the docs does the term appear?
			for doc_id in subtable.keys():
				occurences_in_this_doc = 0
				if self.phrase_search:
					occurences_in_this_doc = len(subtable[doc_id])
				else:
					#print subtable[doc_id]
					if type(subtable[doc_id]) is types.TupleType:
						occurences_in_this_doc = subtable[doc_id][1]
					else:
						occurences_in_this_doc = subtable[doc_id]
				#print "%s: occ=%s" % (term.encode('latin1'), occurences_in_this_doc)
				# Weight as in Perlfect Search:
				#weight = occurences_in_this_doc * math.log(docs_count/term_count)
				# Take into account how long the document is. An occurence in a long
				# document isn't as relevant as one in a short document:
				term_freq = occurences_in_this_doc/float(self.docs_table[doc_id].term_count)
				global_freq = docs_count/float(term_count)
				weight = term_freq * math.log(global_freq) * 100
				#print
				#print "%d,%s: weight = %f * math.log(%f)" % (doc_id, term.encode('latin1'), term_freq, global_freq)
				#print "term_freq = %s/%s" % (occurences_in_this_doc, float(self.docs_table[doc_id].term_count))
				#print subtable[doc_id]
				# TODO?: using an integer for weight makes the index 20% smaller
				#weight = "%.2f" % weight
				if self.phrase_search:
					if type(subtable[doc_id]) is types.TupleType:
						# there's already a tuple if we're updating
						# an existing index.
						#print "X=%s" % str(subtable[doc_id])
						subtable[doc_id] = (weight, subtable[doc_id][1])
					else:
						# new index: create the tuple to store the weight:
						subtable[doc_id] = (weight, subtable[doc_id])
				else:
					subtable[doc_id] = (weight, occurences_in_this_doc)
				self.table[term] = subtable		# required by beedict!
		print ""
		return

	#def merge(self, index2):
	#	"""Merge a second index into this one.
	#	fixme: what if Ids collide?
	#	fixme: update ranking"""
	#	for term2 in index2.table.keys():
	#		if self.table.has_key(term2):
	#			subtable2 = index2.table[term2]
	#			subtable = self.table[term2]
	#			for doc_id in subtable2.keys():
	#				if subtable.has_key(doc_id):
	#					print "*** Error: ID collision doc_id = %s!" % doc_id
	#				else:
	#					subtable[doc_id] = subtable2[doc_id]
	#		else:
	#			self.table[term2] = index2.table[term2]
	#	return
