#!/usr/bin/python
# This file is part of DesktopDig
# Copyright (C) 2002 Daniel Naber <daniel.naber@t-online.de>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import os 
import re
import sys

import FullTextIndex
import Tokenizer

import unittest

class FullTextIndexTest(unittest.TestCase):

	list_bla_blubb = [("bla", [(1, [0])]), ("blubb", [(1, [1])])]
	list_xxx_blubb = [("bla", [(1, [0])]), ("blubb", [(1,[1]),(2,[1])]), ("xxx", [(2,[0])])]
	
	def testAdd(self):
		idx = FullTextIndex.FullTextIndex(basedir, "backend/test_db/one_test", 0)
		idx.deleteData()
		idx.bindData()
		assert(idx.toList() == [])
		idx.addDocument(None, "bla blubb")
		idx.finishIndex()
		self.assertEqual(idx.toList(), self.list_bla_blubb)
		idx.addDocument(None, "xxx blubb")
		idx.finishIndex()
		self.assertEqual(idx.toList(), self.list_xxx_blubb)
		# adding twice? doesn't harm (yet):
		# now it doubles the positions :-(  -- FIXME
		#idx.addDocument(None, "xxx blubb")
		#print "###############"+str(idx.toList())
		#assert(idx.toList() == [("bla", [(1, [0])]), ("blubb", [(1,[1]),(2,[1])]), ("xxx", [(2,[0])])])
		return

	def testAddMultipleOccurences(self):
		# todo: no code duplication!
		idx = FullTextIndex.FullTextIndex(basedir, "backend/test_db/one_test", 0)
		idx.deleteData()
		idx.bindData()
		idx.finishIndex()
		self.assertEqual(idx.toList(), [])
		idx.addDocument(None, "xxx blubb blubb daddel blubb")
		idx.addDocument(None, "blubb blubb")
		idx.finishIndex()
		self.assertEqual(idx.toList(), [("blubb", [(1,[1,2,4]),(2,[0,1])]), ("daddel", [(1,[3])]), ("xxx", [(1,[0])])])
		
	def testLoadAndSave(self):
		idx = FullTextIndex.FullTextIndex(basedir, "backend/test_db/one_test", 0)
		idx.deleteData()
		idx.bindData()
		idx.addDocument(None, "bla blubb")
		idx.addDocument(None, "xxx blubb")
		idx.finishIndex()
		expected_list = [("bla", [(1,[0])]), ("blubb", [(1,[1]),(2,[1])]), ("xxx", [(2,[0])])]
		self.assertEqual(idx.toList(), expected_list)
		# "save":
		idx.commitData()

		# "load":
		idx2 = FullTextIndex.FullTextIndex(basedir, "backend/test_db/one_test", 1)
		idx2.bindData()
		self.assertEqual(idx2.toList(), expected_list)
		return

	def testRemoveDocument(self):
		# todo: no code duplication!
		idx = FullTextIndex.FullTextIndex(basedir, "backend/test_db/one_test", 0)
		idx.deleteData()
		idx.bindData()
		idx.addDocument(None, "bla blubb")
		idx.addDocument(None, "xxx blubb")
		idx.finishIndex()
		self.assertEqual(idx.toList(), [("bla", [(1,[0])]), ("blubb", [(1,[1]),(2,[1])]), ("xxx", [(2,[0])])])
		idx.removeDocument(2)
		# there's no cleanup -> [] stays in the list:
		idx.finishIndex()
		self.assertEqual(idx.toList(), [("bla", [(1,[0])]), ("blubb", [(1,[1])]), ("xxx", [])])
		# deleting twice? doesn't harm (yet). (fixme: expect exception?)
		#idx.removeDocument(2)
		#idx.finishIndex()
		self.assertEqual(idx.toList(), [("bla", [(1,[0])]), ("blubb", [(1,[1])]), ("xxx", [])])
		idx.removeDocument(1)
		idx.finishIndex()
		self.assertEqual(idx.toList(), [("bla", []), ("blubb", []), ("xxx", [])])
		return

	def testUpdateDocument(self):
		# todo: no code duplication!
		idx = FullTextIndex.FullTextIndex(basedir, "backend/test_db/one_test", 0)
		idx.deleteData()
		idx.bindData()
		idx.addDocument(None, "bla blubb")
		idx.addDocument(None, "xxx blubb")
		idx.finishIndex()
		idx.updateDocument(2, "blubb hurz")
		idx.finishIndex()
		self.assertEqual(idx.toList(), [("bla", [(1,[0])]), ("blubb", [(1,[1]),(2,[0])]), ("hurz", [(2,[1])]), ("xxx", [])])
		idx.updateDocument(1, "bla blubb")	# no change
		idx.finishIndex()
		self.assertEqual(idx.toList(), [("bla", [(1,[0])]), ("blubb", [(1,[1]),(2,[0])]), ("hurz", [(2,[1])]), ("xxx", [])])
		idx.updateDocument(1, "muckefuck")
		idx.finishIndex()
		self.assertEqual(idx.toList(), [("bla", []), ("blubb", [(2,[0])]), ("hurz", [(2,[1])]), ("muckefuck", [(1,[0])]), ("xxx", [])])
		idx.updateDocument(1, "")
		idx.finishIndex()
		self.assertEqual(idx.toList(), [("bla", []), ("blubb", [(2,[0])]), ("hurz", [(2,[1])]), ("muckefuck", []), ("xxx", [])])
		return

	def testSearchTerm(self):
		# todo: no code duplication!
		idx = FullTextIndex.FullTextIndex(basedir, "backend/test_db/one_test", 0)
		idx.deleteData()
		idx.bindData()
		idx.addDocument(None, "bla blubb")
		idx.addDocument(None, "xxx blubb")
		idx.finishIndex()
		idx.calculateRanking()

		self.assertEqual(idx.searchTerm("bla"), [(1, 34.657359027997266)])
		# relevancy = 0  because term appears in all (two) documents:
		self.assertEqual(idx.searchTerm("blubb"), [(1,0),(2,0)])

		self.assertEqual(idx.searchTerm("ablubb"), [])
		self.assertEqual(idx.searchTerm("blubbs"), [])
		self.assertEqual(idx.searchTerm("nono"), [])

	def splitAtSpace(self, s):
		l = re.compile("\s+").split(s)
		return l
		
	def testSearchPhrase(self):
		# todo: no code duplication!
		idx = FullTextIndex.FullTextIndex(basedir, "backend/test_db/one_test", 0)
		idx.deleteData()
		idx.bindData()
		idx.addDocument(None, "bla blubb")
		idx.addDocument(None, "xxx blubb")
		idx.addDocument(None, "eins zwei drei vier")
		idx.finishIndex()
		idx.calculateRanking()
		
		# Phrase search without the position part, i.e. AND search in fact:
		#self.assertEqual(idx.searchPhrase("bla blubb"), [(1, 1.0)])
		#self.assertEqual(idx.searchPhrase("blax blubb"), [])
		#self.assertEqual(idx.searchPhrase("blax blubbx"), [])
		#self.assertEqual(idx.searchPhrase("xxx blubb"), [(2, 1.0)])
		#self.assertEqual(idx.searchPhrase("blubb xxx"), [(2, 1.0)])
		#self.assertEqual(idx.searchPhrase("blubb aaa xxx"), [])

		self.assertEqual(idx.searchPhrase(self.splitAtSpace("nono")), [])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("blubb bla")), [])
		
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("bla blubb")), [(1, 75.203869838813702)])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("bla blubb bla")), [])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("blubb bla")), [])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("yyy blubb bla")), [])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("yyy blubb")), [])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("blubb yyy")), [])

		self.assertEqual(idx.searchPhrase(self.splitAtSpace("xxx blubb")), [(2, 75.203869838813702)])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("blubb xxx")), [])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("xxx blubb yyy")), [])

		self.assertEqual(idx.searchPhrase(self.splitAtSpace("eins zwei drei")), [(3, 82.395921650108235)])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("eins zwei drei vier")), [(3, 109.86122886681098)])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("zwei drei")), [(3, 54.930614433405488)])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("drei vier")), [(3, 54.930614433405488)])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("zwei drei vier")), [(3, 82.395921650108235)])

		self.assertEqual(idx.searchPhrase(self.splitAtSpace("zwei dreix vier")), [])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("zweix drei vier")), [])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("zwei drei vierx")), [])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("zwei eins drei vier")), [])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("vier drei zwei eins")), [])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("zwei eins drei")), [])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("zwei eins drei vier")), [])
		self.assertEqual(idx.searchPhrase(self.splitAtSpace("eins xx zwei drei")), [])

		return

	# TODO...
	#def testMerge(self):
	#	idx = FullTextIndex.FullTextIndex(basedir, "backend/test_db/one_test", 0)
	#	idx.deleteData()
	#	idx.bindData()
	#	idx.addDocument(None, "bla blubb blubb")
	#	idx.addDocument(None, "xxx")
	#	idx.finishIndex()
	#
	#	# FIXME: how to handle the ID collision?!?!
	#	idx2 = FullTextIndex.FullTextIndex(basedir, "backend/test_db/two_test", 0)
	#	idx2.deleteData()
	#	idx2.bindData()
	#	idx2.addDocument(None, "bla murks")
	#	idx2.finishIndex()
	#	self.assertEqual(idx2.toList(), [("bla", [(1,[0])]), ("murks", [(1,[1])])])
	#
	#	idx.merge(idx2)
	#	idx.finishIndex()
	#	self.assertEqual(idx.toList(), [('bla', [(1,[0]),(3,[0])]), ('blubb', [(1,[1,2])]), ('murks', [(3,[1])]), ('xxx', [(2,[0])])])
	#	return
	
# the base directory is always the top directory (containing "index.py" etc.):
basedir = os.path.abspath(os.path.split(sys.argv[0])[0])
basedir_match = re.compile("(.*)/.*?").match(basedir)
basedir = basedir_match.group(1)

if __name__ == "__main__":
    unittest.main()
