# This file is part of DesktopDig
# Copyright (C) 2002 Daniel Naber <daniel.naber@t-online.de>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import os
import random
import re
import stat
import time

import Converter
import FullTextIndex

class Indexer:

	def __init__(self, basedir, update, suffixes, desc_words, 
			min_length, max_length, phrase_search, 
			default_encoding, files_world_readable, 
			app=None, progress_dialog=None):
		self.basedir = basedir
		self.update = update
		self.sizecount = 0
		self.rec_count = 0
		self.suffixes = suffixes 	# use None as one list element to index files without suffix
		self.desc_words = desc_words
		self.min_length = min_length
		self.max_length = max_length
		self.phrase_search = phrase_search
		self.default_encoding = default_encoding
		self.files_world_readable = files_world_readable
		self.app = app
		self.progress_dialog = progress_dialog
		self.files_per_directory = {}
		self.abort = 0
		self.dir_sizes = 0
		return
	
	def start(self, directories, index_name):
		self.idx = FullTextIndex.FullTextIndex(self.basedir, index_name, \
			searchonly=0, min_length=self.min_length, 
			max_length=self.max_length, phrase_search=self.phrase_search,
			files_world_readable=self.files_world_readable)
		if not self.update:
			self.idx.deleteData()
		self.idx.bindData()
		for directory in directories:
			if self.abort:
				break
			self.indexDirectories(directory, top_dir=directory)
		if not self.abort:
			# fixme?: can we just leave this out on abort?
			self.idx.finishIndex()
			if self.update:
				self.idx.cleanupIndex()
			print "Calculating ranking..."
			self.idx.calculateRanking()
		# term occurences:
		#for k in self.idx.words_occ_table.keys():
		#	print "%s: %d" % (k, self.idx.words_occ_table[k])
		#print self.idx
		self.idx.commitData()
		#pickle.dump(self.idx, open("/home/dnaber/diplom/dump_file", 'w'))
		# make some statistics:
		information = {}
		information['files'] = self.idx.count
		information['files_size'] = self.sizecount
		information['average_file_size'] = 0
		if self.idx.count > 0:
			information['average_file_size'] = (self.sizecount/1024)/float(self.idx.count)
		#print "Time spent in addDocumentToIndex() = %.2fs" % self.idx.timect
		return information

	def acceptFilename(self, filename):
		"""Should we index this file, according to its suffix?"""
		suffix = re.compile(".*\.(.*)").match(os.path.basename(filename))
		if (suffix and suffix.group(1).lower() in self.suffixes) or \
			((not suffix) and (None in self.suffixes)):
			return 1
		else:
			return 0

	def getDirectoriesSize(self, directory):
		print "getDirectoriesSize(): Entering %s..." % directory
		dir_content = os.listdir(directory)
		for filename in dir_content:
			if self.app:
				self.app.processEvents()
			filename = os.path.join(directory, filename)
			if os.path.isdir(filename):
				self.getDirectoriesSize(filename)
			elif os.path.isfile(filename):
				if self.acceptFilename(filename):
					conv = Converter.Converter(filename, self.desc_words, self.default_encoding)
					size = conv.getDocumentSize()
					#print "file= %s, %d" % (filename, size)
					self.dir_sizes = self.dir_sizes + size
		return self.dir_sizes
		
	def indexDirectories(self, directory, top_dir):
		"""Recursively index all files in this directory and in the
		directories below."""
		print "Entering %s..." % directory
		# top_dir = one of the top directories, as configured by the user
		# count the dirs as added by user, not each dir recursivly:
		if not self.files_per_directory.has_key(top_dir):
			self.files_per_directory[top_dir] = 0
		self.rec_count = self.rec_count + 1
		#print "depth=%d" % 	self.rec_count
		dir_content = os.listdir(directory)
		for filename in dir_content:
			# TODO: this is still not often enough:
			if self.app:
				self.app.processEvents()
			if self.progress_dialog:
				if self.progress_dialog.wasCancelled():
					# fixme? can we just stop here?
					# fixme: this blocks very long?!?!
					self.abort = 1
					return
				self.progress_dialog.setProgress(self.sizecount)

			filename = os.path.join(directory, filename)
			if os.path.isfile(filename):
				if self.acceptFilename(filename):
					filename = os.path.abspath(filename)
					conv = Converter.Converter(filename, self.desc_words, self.default_encoding)
					try:
						(err, filesize, date, title, author, desc, contents) = conv.getDocumentInfo()
					except LookupError, e:
						err = "Decoding error: %s" % e
					except UnicodeError, e:
						err = "Unicode error: %s" % e
					
					if err:
						print "\t-: '%s'" % filename
						print err
						continue
					
					update_needed = None
					doc_id = None
					if self.update:
						(update_needed, doc_id) = self.updateNeeded(filename)
					if self.update and update_needed and doc_id != -1:
						self.files_per_directory[top_dir] += 1
						print "\t%d: updating '%s' (updating) (%.1f kb)" % (doc_id, filename, filesize/1024.0)
						#t1 = time.time()
						self.idx.updateDocument(doc_id, contents, title, author, desc, 
							filename, filesize, date)
						#print "time = %.2f" % (time.time()-t1)
						self.sizecount = self.sizecount + filesize
					elif self.update and (not update_needed and doc_id != -1):
						self.files_per_directory[top_dir] += 1
						print "\t   ignoring '%s' (no update needed) (%.1f kb)" % (filename, filesize/1024.0)
						pass
					else:
						self.files_per_directory[top_dir] += 1
						doc_id = self.idx.addDocument(None, contents, \
							title, author, desc, filename, filesize, date)
						print "\t%d: added    '%s' (%.1f kb)" % (doc_id, filename, filesize/1024.0)
						self.sizecount = self.sizecount + filesize
			elif os.path.isdir(filename):
				self.indexDirectories(filename, top_dir)
			else:
				print "Ignoring %s (link?)" % filename
		self.rec_count = self.rec_count - 1
		#print "self.sizecount=%d" % self.sizecount
		return

	def updateNeeded(self, filename):
		"""File existed when the latest index was made and has been
		changed since then."""
		update_needed = 0
		# to test the update speed: ----
		#if int(random.random()*100) % 5 == 0:
		#	return 1
		#else:
		#	return 0
		# ------------------------------
		prev_time = -1
		# TODO: we'll need yet another index to get this fast...
		filename = os.path.abspath(filename)
		doc_id = None
		for doc_id_tmp in self.idx.docs_table.keys():
			if self.idx.docs_table[doc_id_tmp].url == filename:
				#print "XXX=%s" % self.idx.docs_table[doc_id_tmp].url
				prev_time = self.idx.docs_table[doc_id_tmp].date
				doc_id = doc_id_tmp
				break
		if prev_time == -1:
			# the file is new, didn't exist last time
			#print "*** Warning: unknown last modified date for '%s'" % filename
			#self.idx.docs_table[doc_id].date = os.stat(filename).st_mtime
			update_needed = 0
			doc_id = -1		# = "new file"
		elif os.stat(filename).st_mtime > prev_time:
			#print "### %d > %d (doc_id=%d)" % (os.stat(filename).st_mtime, prev_time, doc_id)
			update_needed = 1
		else:
			update_needed = 0
		return (update_needed, doc_id)
