# This file is part of DesktopDig
# Copyright (C) 2002 Daniel Naber <daniel.naber@t-online.de>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import math
import re
import string
import types	
import IgnoreTerms
import FullTextIndex
import Tokenizer

class Search:

	def __init__(self, basedir, query_terms, index_name, min_length, 
			max_length, phrase_search):
		self.basedir = basedir
		self.query_terms = query_terms
		self.index_name = index_name
		self.min_length = min_length
		self.max_length = max_length
		self.phrase_search = phrase_search
		(self.query_tokens, self.ignored_tokens) = self.getQueryTokens()
		self.idx = FullTextIndex.FullTextIndex(self.basedir, \
			self.index_name, searchonly=1, min_length=min_length, 
			max_length=max_length, phrase_search=phrase_search)
		self.idx.bindData()
		return

	def setSearchTerms(self, terms):
		self.query_terms = terms
		(self.query_tokens, self.ignored_tokens) = self.getQueryTokens()
		return
	
	def getQueryTokens(self):
		tokenizer = Tokenizer.Tokenizer()
		tokens = tokenizer.tokenize(self.query_terms)
		ignore_terms = IgnoreTerms.IgnoreTerms(self.basedir, self.min_length, self.max_length)
		(new_tokens, ignored_tokens) = self.cleanList(tokens, ignore_terms)
		return (new_tokens, ignored_tokens)
		
	def cleanList(self, tokens, ignore_terms):
		i = 0
		new_tokens = []
		ignored_tokens = []
		for term in tokens:
			if type(term) is types.ListType:
				(s_list, s_ignore) = self.cleanList(term, ignore_terms)
				new_tokens.append(s_list)
				if s_ignore:
					ignored_tokens.extend(s_ignore)
			else:
				new_term = ignore_terms.acceptTerm(term)
				if new_term:
					new_tokens.append(new_term)
				else:
					ignored_tokens.append(term)
			i = i + 1
		return (new_tokens, ignored_tokens)

	def highlightTerms(self, s):
		for term in self.query_tokens:
			regex = re.compile('\\b(%s)\\b' % term, re.IGNORECASE)
			m = regex.search(s)
			if m:
				s = regex.sub('<strong>%s</strong>' % m.group(1), s)
		return s
		
	def countDocuments(self):
		return self.idx.countDocuments()

	def getAndResults(self):
		"""A search that matches only documents which contain *all* searched terms."""
		match_set = {}
		i = 0
		for token in self.query_tokens:
			if type(token) is types.ListType:
				matches = self.idx.searchPhrase(token)
			else:
				matches = self.idx.searchTerm(token)
			new_matches = {}
			for doc_id, relev in matches:
				new_matches[doc_id] = relev
			if i == 0:	# first token
				for result_id, relevancy in matches:
					match_set[result_id] = relevancy
			else:		# all other tokens
				for result_id in match_set.keys():
					try:
						new_relev = new_matches[result_id]
						match_set[result_id] = match_set[result_id] + new_relev
					except KeyError:
						del match_set[result_id]
			i = i + 1
		matches = []
		for match_id in match_set.keys():
			if self.phrase_search:
				relevancy = match_set[match_id]
			else:
				relevancy = match_set[match_id][0]
			doc = self.idx.docs_table[match_id]
			matches.append(Match(doc, relevancy))
			#print "%s-%f" % (doc, relevancy)
		matches.sort()
		return matches

	def getOrResults(self):
		"""A search that matches all documents which contain at least one searched terms."""
		boost = 1	# give more weight to documents that contain all or many of the search terms
				# todo: should we instead use an algorithm that *guarantees* that these
				# documents appears at the top?
		match_set = {}
		for token in self.query_tokens:
			if type(token) is types.ListType:
				matches = self.idx.searchPhrase(token)
			else:
				matches = self.idx.searchTerm(token)
			for doc_id, relev_tuple in matches:
				if self.phrase_search:
					relev = relev_tuple
				else:
					relev = relev_tuple[0]
				try:
					matched_terms, relev = match_set[doc_id]
					new_relev = match_set[doc_id][0] + relev
					match_set[doc_id] = (matched_terms + 1, new_relev)
				except KeyError:
					# as the first value of the tuple save the number of matching terms
					# (obviously this is between 1 and len(self.query_tokens))
					match_set[doc_id] = (1, relev)
		matches = []
		for match_id in match_set.keys():
			matched_terms, relevancy = match_set[match_id]
			doc = self.idx.docs_table[match_id]
			if boost:
				erg = matched_terms / float(len(self.query_tokens))
				# we enhance the usual ranking: if more than one
				# term matches for a document, this will move the 
				# matched document up in the ranking:
				if erg >= 1:
					erg = 0.95
				factor = abs(1/math.log(erg))
				relevancy = relevancy * factor 
				#print "**%d (%s): erg=%.2f factor=%.2f" % (match_id, doc.title, erg, factor)
			matches.append(Match(doc, relevancy))
		matches.sort()
		return matches

class Match:
	"""A doc and its relevancy (wrt. a certain query, which is not part of this object).
	Can be sorted by relevancy, most relevant first."""
	
	def __init__(self, doc_object, relevancy):
		self.doc = doc_object
		self.relevancy = relevancy
		return

	def __cmp__(self, that):
		if self.relevancy < that.relevancy:
			return 1
		elif self.relevancy > that.relevancy:
			return -1
		else:
			# We want a stable sort order, so use the url as a second sort key:
			if self.doc.url < that.doc.url:
				return 1
			elif self.doc.url > that.doc.url:
				return -1
			else:
				# Guarantee stable sort order for the test cases, so use
				# the title as a third sort key (since the url is the same for all test cases):
				if self.doc.title < that.doc.url:
					return 1
				else:
					return -1
