#!/usr/bin/python
# This file is part of DesktopDig
# Copyright (C) 2002,2003 Daniel Naber <daniel.naber@t-online.de>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import os
import re
import string
import sys
import time

from qt import *
from kdeui import *
from kdecore import KApplication, KCmdLineArgs, KURL
#from kio import KDirWatch,  KMimeType, KServiceTypeProfile
from kio import KMimeType, KRun
from kfile import KFileDialog

import Config
import Search
import Indexer
import FullTextIndex

true = 1
false = 0

class MyQTextBrowser(QTextBrowser):

	def __init__(self, *args):
		apply(QTextBrowser.__init__, (self,) + args)
		return

class ConfigDialog(KDialogBase):

	margin = 4
	spacing = 2

	def __init__(self, suffixes, min_len_val, max_len_val, desc_len_val):
		apply(KDialogBase.__init__, (self,) + (KDialogBase.Swallow, "Global Configuration", \
			KDialogBase.Ok|KDialogBase.Cancel, KDialogBase.Ok))
		page = QWidget(self)
		self.setMainWidget(page)
		topLayout = QVBoxLayout(page, 0, self.spacing)
		
		label = QLabel("Index files with these suffixes (comma separated list):", \
			page, "suffixes")
		topLayout.addWidget(label)
		
		suffix_none = 0
		suffixes_new = []
		for suffix in suffixes:
			if not suffix:
				suffix_none = 1
			else:
				suffixes_new.append(suffix)
		suffixes_str = string.join(suffixes_new, ', ')
		self.lineedit = QLineEdit(suffixes_str, page, "lineedit")
		topLayout.addWidget(self.lineedit)
		
		self.no_suffix_check = QCheckBox("Index file with no suffix (e.g. \"README\")", page)
		self.no_suffix_check.setChecked(suffix_none)
		topLayout.addWidget(self.no_suffix_check)

		grid_layout = QGridLayout(page, 3, 2, self.margin, self.spacing)

		self.min_len = QLineEdit(str(min_len_val), page)
		label = QLabel(self.min_len, "Minimum word length:", page)
		grid_layout.addWidget(label, 0, 0)
		grid_layout.addWidget(self.min_len, 0, 1)

		self.max_len = QLineEdit(str(max_len_val), page)
		label = QLabel(self.max_len, "Maximum word length:", page)
		grid_layout.addWidget(label, 1, 0)
		grid_layout.addWidget(self.max_len, 1, 1)

		self.desc_len = QLineEdit(str(desc_len_val), page)
		label = QLabel(self.desc_len, "Words saved as description:", page)
		grid_layout.addWidget(label, 2, 0)
		grid_layout.addWidget(self.desc_len, 2, 1)

		topLayout.addLayout(grid_layout)
		
		return
		
	def getSuffixes(self):
		suffixes = str(self.lineedit.text())
		suffixes_list_tmp = re.compile(",\s*").split(suffixes)
		suffixes_list = []
		for item in suffixes_list_tmp:
			# remove empty items
			if item.strip():
				suffixes_list.append(item)
		if self.no_suffix_check.isChecked():
			suffixes_list.append(None)
		return suffixes_list

	def getMinLength(self):
		return int(str(self.min_len.text()))

	def getMaxLength(self):
		return int(str(self.max_len.text()))

	def getDescLength(self):
		return int(str(self.desc_len.text()))
			
class MainWindow(KMainWindow):

	margin = 4
	spacing = 2
	
	def __init__(self, *args):
		#apply(KMainWindow.__init__, (self,) + args)
		apply(KMainWindow.__init__, (self,))
		self.setCaption("DesktopDig")
		self.resize(500, 400)

		m_tab = QTabWidget(self)
		self.main_vb = QVBoxLayout(self, self.margin, -1, "self.main_vb")
		self.main_vb.addWidget(m_tab)

		self.searchWidget = self.setupSearchTab()
		m_tab.addTab(self.searchWidget, "&Search")
		self.indexingWidget = self.setupIndexTab()
		m_tab.addTab(self.indexingWidget, "&Index")

		self.button_box = QHBoxLayout(self, self.margin, -1, "self.button_box")
		self.quit_button = QPushButton("&Quit", self)
		self.button_box.addWidget(self.quit_button)
		QObject.connect(self.quit_button, SIGNAL("clicked()"), self.slotQuit)

		self.hide_button = QPushButton("&Hide", self)
		self.button_box.addWidget(self.hide_button)
		QObject.connect(self.hide_button, SIGNAL("clicked()"), self.slotHide)
		self.main_vb.addLayout(self.button_box)
		
		self.setCentralWidget(m_tab)
		QObject.connect(m_tab, SIGNAL("currentChanged(QWidget *)"), self.slotTabSelected)
		#self.statusBar().insertItem("x", 0)	 fixme, doesn't work

		#### System tray applet:
		self.tray = KSystemTray(self)
		pixmappath = os.path.join(Config.BASEDIR, "etc", "dig.png")
		pixmap = QPixmap(pixmappath)
		self.tray.setPixmap(pixmap)
		QToolTip.add(self.tray, "DesktopDig")
		self.tray.show()
	
		### load config:
		self.cfg = Config.Config()
		config_dirs = self.cfg.getList("DIRECTORIES")
		if len(config_dirs) == 1 and config_dirs[0] == '':
			# use a KDE standard directory as default:
			config_dirs[0] = os.path.join(os.environ['HOME'], "Documents")
		index_name = "data/index"
		idx = FullTextIndex.FullTextIndex(Config.BASEDIR,
			index_name, 
			searchonly=0,	# write an index if no one exists yet
			min_length=self.cfg.getInt('MIN_LENGTH'),
			max_length=self.cfg.getInt('MAX_LENGTH'),
			phrase_search=self.cfg.getInt('PHRASE_SEARCH'))
		# TODO: how can this be avoided?:
		idx.bindData()
		indexed_documents = idx.getIndexDocumentNames()
		for config_dir in config_dirs:
			# TODO: what if an index was built manually with ./indexer.pl but
			# we don't know for which directories? currently it will not
			# appear in this list.
			if config_dir == '':
				continue
			indexed_files = 0
			for indexed_doc in indexed_documents:
				if indexed_doc.startswith(config_dir):
					indexed_files = indexed_files + 1
			item = QListViewItem(self.list, config_dir, str(indexed_files))
			self.list.insertItem(item)

		self.list.setShowSortIndicator(1)
		#self.list.setColumnWidthMode(0, )
		self.list.setResizeMode(QListView.AllColumns)
		QObject.connect(self.list, SIGNAL("selectionChanged()"),
			self.slotSelectionChanged)
		QObject.connect(self.list, SIGNAL("doubleClicked(QListViewItem *)"),
			self.slotDoubleClicked)
		
		self.hb.addWidget(self.list)
		
		self.inner_vb = QVBoxLayout(self.hb, self.spacing)
		#self.inner_vb.setDirection(QBoxLayout.TopToBottom)
		self.add_button = QPushButton("Add...", self.indexingWidget)
		QObject.connect(self.add_button, SIGNAL("clicked()"), self.slotAdd)
		self.edit_button = QPushButton("Edit...", self.indexingWidget)
		QObject.connect(self.edit_button, SIGNAL("clicked()"), self.slotEdit)
		self.delete_button = QPushButton("Remove...", self.indexingWidget)
		QObject.connect(self.delete_button, SIGNAL("clicked()"), self.slotDelete)
		self.config_button = QPushButton("Configure...", self.indexingWidget)
		QObject.connect(self.config_button, SIGNAL("clicked()"), self.slotConfigure)
		self.update_button = QPushButton("Update Index", self.indexingWidget)
		QObject.connect(self.update_button, SIGNAL("clicked()"), self.slotUpdate)
		self.index_button = QPushButton("Build Index", self.indexingWidget)
		QObject.connect(self.index_button, SIGNAL("clicked()"), self.slotBuildIndex)
		self.inner_vb.addWidget(self.add_button, 0, Qt.AlignTop)
		self.inner_vb.addWidget(self.edit_button, 0, Qt.AlignTop)
		self.inner_vb.addWidget(self.delete_button, 0, Qt.AlignTop)
		self.inner_vb.addStretch(1)
		self.inner_vb.addWidget(self.config_button, 0, Qt.AlignTop)
		self.inner_vb.addWidget(self.index_button, 0, Qt.AlignTop)
		self.inner_vb.addWidget(self.update_button, 0, Qt.AlignTop)

		self.watcher = None
		return

	def setupSearchTab(self):
		#### Search tab:
		searchWidget = QWidget(self, "searchWidget")
		self.search_vb = QVBoxLayout(searchWidget, self.margin, -1, "self.search_vb")
		# FIXME: why does this produce a run time warning ('...already has a layout')?:
		self.search_hb = QHBoxLayout(searchWidget, self.margin, -1, "self.search_hb")
		self.query = QLineEdit(searchWidget)
		self.query.setFocus()
		self.query_mode = QComboBox(searchWidget)
		self.query_mode.insertItem("All words")
		self.query_mode.insertItem("Any word")
		self.query_submit = QPushButton("Search", searchWidget)
		QObject.connect(self.query, SIGNAL("returnPressed()"),
			self.slotSearch)
		QObject.connect(self.query_submit, SIGNAL("clicked()"), self.slotSearch)
		self.result = MyQTextBrowser(searchWidget)
		# TODO: to remove the hack in linkClicked(), but disconnect() not
		# yet supported by PyQt:
		#self.result.disconnect(SIGNAL("linkClicked(const QString&)"))
		QObject.connect(self.result, SIGNAL("linkClicked(const QString&)"),
			self.slotGotoLink)
		self.query_label = QLabel(self.query, "Search &for:", searchWidget)
		self.search_hb.addWidget(self.query_label, 0)
		self.search_hb.addWidget(self.query, 0)
		self.search_hb.addWidget(self.query_mode, 0)
		self.search_hb.addWidget(self.query_submit, 0)
		self.search_vb.addLayout(self.search_hb, 0)
		self.search_vb.addWidget(self.result, 1)
		return searchWidget

	def setupIndexTab(self):
		#### Indexer tab:
		indexingWidget = QWidget(self)
		self.hb = QHBoxLayout(indexingWidget, self.margin, self.spacing, "hb")
		self.list = QListView(indexingWidget)
		self.list.setAllColumnsShowFocus(true)
		self.list.setColumnAlignment(1, Qt.AlignRight)
		self.list.addColumn("Directory")
		self.list.addColumn("Files Indexed")
		return indexingWidget

	def formatResults(self, matches, search):
		res= []
		res.append('<qt>')
		i = 1
		for match in matches:
			title = match.doc.title
			if not title:
				title = match.doc.url
			# TODO: make matching words bold
			size = match.doc.size/1024
			if size == 0:
				size = 1
			desc = "(No description available)"
			if match.doc.description:
				desc = search.highlightTerms(match.doc.description)
			date = time.strftime(self.cfg.getString('TIME_FORMAT'), time.localtime(match.doc.date))
			res.append('<p><a href="%s">%s</a><br>'
				'%s<br>'
				'<font color="darkgreen">%s - Size:&nbsp;%dk - '
					'Score:&nbsp;%.2f - Date:&nbsp;%s</font>'
				'</p>\n' %
				(match.doc.url, title, desc, match.doc.url, 
				size, match.relevancy, date))
			i = i + 1
		if len(matches) == 0:
			res.append("No matches.")
		res.append("</qt>")
		return string.join(res, '')

	def enableActionButtons(self, enable):
		self.edit_button.setEnabled(enable)
		self.delete_button.setEnabled(enable)
		return


	##### SLOTS:

	def slotTabSelected(self, n):
		if n == self.searchWidget:
			self.query.setFocus()
		else:
			if self.list.selectedItem():
				self.enableActionButtons(1)
			else:
				self.enableActionButtons(0)
		return
		
	def slotSearch(self):
		self.setWaitCursor()
		query = unicode(str(self.query.text().utf8()), 'utf-8')
		#print "Searching '%s'" % query
		index_name = "data/index"
		s = Search.Search(Config.BASEDIR, query, index_name, \
			self.cfg.getInt('MIN_LENGTH'), self.cfg.getInt('MAX_LENGTH'),
			self.cfg.getInt('PHRASE_SEARCH'))
		matches = []
		if self.query_mode.currentItem() == 0:
			# requires that 'AND' is the first item
			matches = s.getAndResults()
		elif self.query_mode.currentItem() == 1:
			matches = s.getOrResults()
		else:
			print "Search type '%s' not supported." % search_type
			return
		#self.statusBar().changeItem("xxx", 0)		# fixme
		res = self.formatResults(matches, s)
		self.result.setText(res)
		self.current_result = res
		self.restoreCursor()
		return

	def slotGotoLink(self, url):
		print "Go to '%s'?" % url
		if not os.path.exists(str(url)):
			KMessageBox.error(self, "This file doesn't exist anymore.", "File not found")
		else:
			choice = KMessageBox.warningContinueCancel(self, 
				"Open this file with its associated application? Warning: \
				This might execute the file directly (e.g. for shell scripts).",
				"Open file?", KStdGuiItem.cont(), "askagain")
			if choice == KMessageBox.Continue:
				self.krun = KRun(KURL(KURL.encode_string(url)), 0, true, 1)
				# fixme?: this seems to fire to early:
				QObject.connect(self.krun, SIGNAL("finished()"), self.restoreCursor)
				QObject.connect(self.krun, SIGNAL("error()"), self.execError)
		# FIXME: this is an ugly hack so the result
		#  doesn't disappear when the user clicks on a link (see "disconnect"):
		self.result.setText(self.current_result)
		return

	def slotSelectionChanged(self):
		if self.list.selectedItem():
			self.enableActionButtons(1)
		else:
			self.enableActionButtons(0)
		return

	def slotDoubleClicked(self):
		self.slotEdit()
		return
		
	def slotAdd(self):
		name = str(KFileDialog.getExistingDirectory())
		if name != '':
			item = QListViewItem(self.list, name, "0")
			self.list.insertItem(item)
		return

	def slotEdit(self):
		name = str(KFileDialog.getExistingDirectory(self.list.currentItem().text(0)))
		if name != '':
			self.list.currentItem().setText(0, name)
			self.list.currentItem().setText(1, "0")
		return

	def slotDelete(self):
		item = self.list.currentItem()
		if item:
			self.list.takeItem(item)
		return

	def slotConfigure(self):
		dlg = ConfigDialog(self.cfg.getList('SUFFIXES'), \
			self.cfg.getInt('MIN_LENGTH'), self.cfg.getInt('MAX_LENGTH'), \
			self.cfg.getInt('DESC_WORDS'))
		dlg.exec_loop()
		if dlg.result() == QDialog.Accepted:
			self.saveConfiguration(dlg.getSuffixes(), dlg.getMinLength(), \
				dlg.getMaxLength(), dlg.getDescLength())
		return

	def slotUpdate(self):
		# TODO: make this work:
		#self.watcher = WatchDirectories(self.getList())
		self.setWaitCursor()
		indexer = Indexer.Indexer(Config.BASEDIR, update=1,
			suffixes=self.cfg.getList('SUFFIXES'), \
			desc_words=self.cfg.getInt('DESC_WORDS'),
			min_length=self.cfg.getInt('MIN_LENGTH'), \
			max_length=self.cfg.getInt('MAX_LENGTH'),
			phrase_search=self.cfg.getInt('PHRASE_SEARCH'), \
			default_encoding=self.cfg.getString('DEFAULT_ENCODING'),
			files_world_readable=self.cfg.getInt('FILES_WORLD_READABLE'))
		# FIXME: 
		#	-make this non-blocking!
		#	-show progress bar and stop button
		indexer.start(self.getList(), index_name="data/index")
		self.updateIndexList(indexer)
		self.restoreCursor()
		return
		
	def updateIndexList(self, indexer):
		# update number of indexed files:
		item = self.list.firstChild()
		while item:
			strng = str(item.text(0))
			item.setText(1, str(indexer.files_per_directory[strng]))
			item = item.itemBelow()
		print "Update end."
		return

	def slotBuildIndex(self):
		self.setWaitCursor()
		#print "Build..."
		progress_dialog = QProgressDialog("Indexing...", "Cancel", 1)
		progress_dialog.setMinimumDuration(500)
		indexer = Indexer.Indexer(Config.BASEDIR, update=0, \
			suffixes=self.cfg.getList('SUFFIXES'),
			desc_words=self.cfg.getInt('DESC_WORDS'), \
			min_length=self.cfg.getInt('MIN_LENGTH'),
			max_length=self.cfg.getInt('MAX_LENGTH'), \
			phrase_search=self.cfg.getInt('PHRASE_SEARCH'), \
			default_encoding=self.cfg.getString('DEFAULT_ENCODING'), \
			files_world_readable=self.cfg.getInt('FILES_WORLD_READABLE'), \
			app=app, progress_dialog=progress_dialog)
		# Find out how many files we are going to index. this can take quite some
		# time already, so FIXME: show progress dialog here (with unknown max
		# progress but cancel button):
		max_progress = 0
		for directory in self.getList():
			max_progress += indexer.getDirectoriesSize(directory)
			#print "max_progress = %d" % max_progress
		self.restoreCursor()
		progress_dialog.setTotalSteps(max_progress)
		indexer.start(self.getList(), index_name="data/index")
		progress_dialog.setProgress(max_progress)
		print "Build end."
		self.updateIndexList(indexer)
		return

	def slotQuit(self):
		self.saveDirectories()
		app.quit()
		return

	def slotHide(self):
		self.hide()
		return

	##### end slots

	def queryExit(self):
		self.saveDirectories()
		return 1
		
	def saveConfiguration(self, suffixes=None, min_len=None, max_len=None,
			desc_len=None):
		dict = {}
		if suffixes:
			suffixes_tmp = []
			for s in suffixes:
				if not s:
					suffixes_tmp.append('None')
				else:
					suffixes_tmp.append(s)
			dict['SUFFIXES'] = string.join(suffixes_tmp, ",")
		if min_len:
			dict['MIN_LENGTH'] = min_len
		if max_len:
			dict['MAX_LENGTH'] = max_len
		if desc_len:
			dict['DESC_WORDS'] = desc_len
		self.cfg.save(dict)
		return

	def saveDirectories(self):
		dict = {}
		dict['DIRECTORIES'] = string.join(self.getList(), ",")
		self.cfg.save(dict)
		return

	def setWaitCursor(self):
		app.setOverrideCursor(KCursor.waitCursor());
		return
	
	def execError(self):
		self.restoreCursor()
		KMessageBox.error(self, "The file could not be opened.", "Error opening file")
		return

	def restoreCursor(self):
		app.restoreOverrideCursor()
		return

	def getList(self):
		"""Get a list of directories which should be indexed."""
		item = self.list.firstChild()
		l = []
		while item:
			st = str(item.text(0))
			l.append(st)
			item = item.itemBelow()
		return l

def main():
	main_win = MainWindow()
	main_win.show()		# - we're already visible in the systray
	#app.connect(app, SIGNAL("lastWindowClosed()"), main_win.slotHide)
	app.exec_loop()
	return

appname     = "DesktopDig"
description = "A local search engine for your personal files"
version     = "0.1"
KCmdLineArgs.init(sys.argv, appname, description, version)
app = KApplication()
main()
